// $Id: basefile.h,v 1.10 2005/11/10 15:53:22 troein Exp $
//
// BioArray Software Environment (BASE) - homepage http://base.thep.lu.se/
// Copyright (C) 2002,2003 Lao Saal, Carl Troein, Johan Vallon-Christersson
//
// This file is part of BASE.
//
// BASE is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// BASE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
#ifndef _BASEFILE_H
#define _BASEFILE_H

#ifdef HAVE_CONFIG_H
	#include <config.h>
#endif
#include <fstream>
#include <map>
#include <string>
#include <vector>
#include <cmath>
#include <cstdlib>

namespace BASE
{

// A BaseFileReader is created from an istream, e.g. cin, and sections
// are read from it using readSection().
//
// A BaseFileSection stores information about a section of a BASEfile,
// basically just the section type and the headers.
// The findXxxOpt() functions can be used to extract information from
// the headers, which are stored as a map from string to string. With
// findFieldList() you can extract a tab-separated list of e.g. column
// names. The extracted information is stored in a map from string to
// position, and the position of a given string can be found using
// findFieldPos. These functions all exist to reduce the mucking about
// iterators and map.end().
//
// The data after the headers of a section can be extracted using
// readVec(), which takes a reference to a vector of char *, int,
// float, or some other type. The size of this vector must be the
// expected number of tab-separated columns, or an error will be
// generated. Errors are detected with getError(), after some bool-
// returning function has returned false.
//
// There is no class for generating BASEfiles, since it's so easy to
// do 'manually'.

typedef std::map<std::string,int> FieldList;

class BaseFileSection
{
  public:
	typedef std::map<std::string,std::string> header_t;

	BaseFileSection();

	bool findIntOpt(const std::string &name, int &ret);
	bool findStringOpt(const std::string &name, std::string &ret);
	bool findFieldList(const std::string &name, FieldList &mapping);
	static bool findFieldPos(const std::string &name, int &ret,
		FieldList &mapping);

	std::string type;
	int startLine, dataStartLine;
	header_t headers;
};

class BaseFileReader
{
  public:
	// Argument to ctor: the istream to read from (e.g. cin)
	BaseFileReader(std::istream &s, bool changebuf = true);
	~BaseFileReader();

	// The various errors getError() may return. EOR is end-of-record,
	// and shouldn't be considered an error.
	enum
	{
		EOR = 0, FILEERR = 1, BADFORMAT, TOOMANY, TOOFEW, BADTYPE, BADSECTION,
			SEEKERR, BADERROR
	};

	// Read a single section into a BaseFileSection object
	bool readSection(BaseFileSection &sect);

	// Read a single line of data and split it on tabs. The number of
	// elements in the input vector must be the number of columns
	// expected in the file.
	template<class vectype_>
	bool readVec(std::vector<vectype_> &vec)
	{
		if(!readLine())
			return false;
		if(!buf[0])
		{
			_error = EOR;
			return false;
		}
		char *tok = &buf[0];
		char *otok;
		for(unsigned int i = 0; i < vec.size(); i++)
		{
			if(!tok)
			{
				_error = TOOFEW;
				return false;
			}
			otok = strchr(tok, '\t');
			if(otok) *(otok++) = 0;
			VecTranslate(tok, vec[i]);
			tok = otok;
		}
		if(tok)
		{
			_error = TOOMANY;
			return false;
		}
		return true;
	}

	// Reads the entire data part of the section into a vector
	bool readLines(std::vector<std::string> &vec);

	// Error information functions
	int getError() { return _error; }
	int getLine() { return line; }
	const char *errText(int e)
	{
		return errtexts[std::min((int)BADERROR, std::max(0, e))];
	}
	const char *errText() { return errText(_error); }

	// Seeking functions and support.
	struct SeekPosition
	{
		SeekPosition(): pos(0), line(0) {}
		std::streampos pos;
		int line;
	};
	bool tellPos(SeekPosition &spos);
	bool seekPos(const SeekPosition &spos);


  protected:
	int _error;
	bool isFirstSection() { return _firstSection; }
	static const float fnan;
	static const double dnan;

  private:
	bool readHeadersNew(BaseFileSection::header_t &nameValue);
	inline void VecTranslate(char *t, int &v)
	{
		v = std::atoi(t);
	}
	inline void VecTranslate(char *t, float &v)
	{
		if(!*t) v = fnan;
		else
		{
			char *p;
#ifdef HAVE_STRTOF
			v = std::strtof(t, &p);
#else
			v = std::strtod(t, &p);
#endif
			if(*p)
				v = fnan;
		}
	}
	inline void VecTranslate(char *t, double &v)
	{
		if(!*t) v = dnan;
		else
		{
			char *p;
			v = std::strtod(t, &p);
			if(*p)
				v = dnan;
		}
	}
	template<class vectype_>
	inline void VecTranslate(char *t, vectype_ &v)
	{
		v = t;
	}

	bool readLine();
	bool skipData();

	// All these should have names starting with _
	std::istream &str;
	int line;
	bool dataEnd;
	bool _firstSection;
	std::vector<char> buf, inbuf;

	static const int inbuflen;

	static const char *errtexts[];
};

// A small helper function that reads from an istream into a vector.
std::istream &getline(std::istream &i, std::vector<char> &v,
	char delim = '\n');

// Another helper function. This one should _not_ be in basefile.cc
void split(const std::string &s, std::vector<std::string> &v,
	size_t maxp = 0, char c = '\t');


}

#endif
