//
// BioArray Software Environment (BASE) - homepage http://base.thep.lu.se/
// Copyright (C) 2002,2003 Bjrn Samuelsson
//
// This file is part of BASE.
//
// BASE is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// BASE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
#include <iostream>
#include <base_lowess_w.h>

namespace BASE
{

const double lowess_weighted::yw_range_factor = 6.;


static inline double cube(double a)
{
	return a * a * a;
}

double lowess_weighted::cubic_weight_func(double a)
{
	if(fabs(a) < 1.)
		return cube(1. - cube(fabs(a)));
	else
		return 0.;
}

static inline double sqr(double a)
{
	return a * a;
}

double lowess_weighted::quadratic_weight_func(double a)
{
	if(fabs(a) < 1.)
		return sqr(1. - sqr(a));
	else
		return 0.;
}


void lowess_weighted::line_fit(double *x_beg, int npoints, double *y_beg,
	double &k, double &m, double *w_beg)
{
	const int np = npoints;
	if(np <= 0)
	{
		std::cerr << "line_fit called with npoints <= 0\n";
		exit(1);
	}
	
	double x_sum = 0., y_sum = 0., xx_sum = 0., xy_sum = 0., w_sum = 0.;

	for(int i = 0; i < np; i++)
	{
		double w = w_beg[i];
		x_sum += x_beg[i] * w;
		y_sum += y_beg[i] * w;
		xx_sum += x_beg[i] * x_beg[i] * w;
		xy_sum += x_beg[i] * y_beg[i] * w;
		w_sum += w;
	}
	
	if(w_sum <= 0.)
	{
		// Fallback when all weights are extremely small.
		for(int i = 0; i < np; i++)
		{
			x_sum += x_beg[i];
			y_sum += y_beg[i];
			xx_sum += x_beg[i] * x_beg[i];
			xy_sum += x_beg[i] * y_beg[i];
		}
		w_sum = np;
	}

	double denom = w_sum * xx_sum - x_sum * x_sum;
	if(denom != 0.)
	{
		k = (w_sum * xy_sum - x_sum * y_sum) / denom;
		m = (y_sum - k * x_sum) / w_sum;
	}
	else
	{
		k = 0.;
		m = y_sum / w_sum;
	}
}

// fit for one point
// x must be sorted in ascending order
void lowess_weighted::p_lowess(double *x_beg, int npoints, double *y_beg,
	const double &p_x, double &p_y, double *temp_w_beg,
	double *yw_beg, weight_func x_weight_func)
{
	const int np = npoints;
	if(np <= 0)
	{
		std::cerr << "p_lowess called with npoints <= 0\n";
		exit(1);
	}

	double inv_radius = 1. / std::max(p_x - *x_beg, x_beg[np - 1] - p_x);
	for(int i = 0; i < np; i++)
	{
		temp_w_beg[i] = x_weight_func((x_beg[i] - p_x) * inv_radius) *
			yw_beg[i];
	}

	double k, m;
	line_fit(x_beg, npoints, y_beg, k, m, temp_w_beg);

	p_y = k * p_x + m;
}

// robust fit of line segments
// x must be sorted in ascending order
void lowess_weighted::calc(double *x_beg, int npoints, double *y_beg,
	double *weight_beg, double *y_fit_beg, char *point_fitted_beg,
	const double &frac, const int n_iter, const double &delta,
	weight_func x_weight_func, weight_func y_weight_func)
{
	if(npoints == 0)
		return;
	if(npoints < 0)
	{
		std::cerr << "lowess_weighted called with npoints < 0\n";
		exit(1);
	}
	if(npoints == 1)
	{
		std::copy(y_beg, y_beg + npoints, y_fit_beg);
		std::fill(point_fitted_beg, point_fitted_beg + npoints, true);
		return;
	}

	double totw = 0.;
	for(int i = 0; i < npoints; i++)
		totw += weight_beg[i];
	if(totw <= 0.)
	{
		std::cerr << "lowess_weighted called with total weight <= 0\n";
		exit(1);
	}
	const double np_window = totw * frac;

	double w_beg[npoints];
	double temp_beg[npoints];
	std::pair<double, double> temp2_beg[npoints];

	for(int i = 0; i < npoints; i++)
		w_beg[i] = weight_beg[i];

//std::cerr<<"totw: "<<totw<<"\n";
	for(int iter = 0; iter <= n_iter; iter++)
	{
		int window_beg = 0, window_end = 1;
		double window_wsum = weight_beg[0];
		int i, prev_i;

		for(i = 0, prev_i = -1; prev_i < npoints - 1;)
		{
			//center window and make it big enough.
			while(window_end < npoints && 
				(x_beg[i] - x_beg[window_beg] > 
				x_beg[window_end] - x_beg[i] ||
				window_wsum < np_window))
			{
				window_wsum += weight_beg[window_end];
				window_end++;
//std::cerr<<"ext to "<<window_beg<<" - "<<window_end<<"   wsum: "<<window_wsum<<"\n";
				while(window_beg < window_end - 1 &&
					window_wsum - weight_beg[window_beg] >= np_window)
				{
					window_wsum -= weight_beg[window_beg];
					window_beg++;
//std::cerr<<" cut to "<<window_beg<<" - "<<window_end<<"   wsum: "<<window_wsum<<"\n";
				}
			}

//if(!iter)	std::cerr<<x_beg[i]<<" "<<x_beg[window_beg]<<" "<<x_beg[window_end-1]<<"\n";

			int w_modpos = -1;
			double old_w_mod = 0;
			if(window_wsum > np_window && window_end > window_beg + 1)
			{
				if(x_beg[i] - x_beg[window_beg] >
					x_beg[window_end - 1] - x_beg[i])
				{
					w_modpos = window_beg;
				}
				else
					w_modpos = window_end - 1;
				old_w_mod = w_beg[w_modpos];
				w_beg[w_modpos] *= (window_wsum / np_window - 1.) / weight_beg[i];
//std::cerr<<"changed "<<old_w_mod<<" to "<<w_beg[w_modpos]<<" at "<<w_modpos<<std::endl;
				if(w_beg[w_modpos] < 0)
				{
					std::cerr << "broken window correction in lowess_weighed\n";
					exit(1);
				}
			}

			//fit one point
			p_lowess(x_beg + window_beg, window_end - window_beg,
				y_beg + window_beg, x_beg[i], y_fit_beg[i], temp_beg,
				w_beg + window_beg, x_weight_func);

			if(w_modpos >= 0)
				w_beg[w_modpos] = old_w_mod;

			//interpolate skipped points
			if(prev_i + 1 < i)
			{
				double d = x_beg[i] - x_beg[prev_i];
				if(d == 0.)
				{
					for(int j = prev_i + 1; j < i; j++)
						y_fit_beg[j] = y_fit_beg[i];
				}
				else
				{
					for(int j = prev_i + 1; j < i; j++)
					{
						double a = (x_beg[j] - x_beg[prev_i]) / d;
						y_fit_beg[j] = a * y_fit_beg[i] +
							(1. - a) * y_fit_beg[prev_i];
					}
				}
			}

			if(point_fitted_beg)
			{
				for(int j = prev_i + 1; j < i; j++)
					point_fitted_beg[j] = false;
				point_fitted_beg[i] = true;
			}

			prev_i = i;
			double cut = x_beg[i] + delta;
			for(i = prev_i + 1; i < npoints && x_beg[i] <= cut; i++)
			{
				if(x_beg[i] - x_beg[prev_i] < 0)
				{
					std::cerr << "x-values not sorted in lowess_weighted\n";
					exit(1);
				}
			}
			if(i >= npoints)
				i = npoints - 1;
		}

		if(iter == n_iter)
			break;
		
		// Find median and reweight points based on y
		for(int i = 0; i < npoints; i++)
		{
			temp2_beg[i] = std::make_pair(fabs(y_fit_beg[i] - y_beg[i]),
				weight_beg[i]);
		}
		std::sort(temp2_beg, temp2_beg + npoints);
		double w = 0.;
		int midpos = 0;
		while(midpos < npoints && w < .5 * totw)
		{
			w += temp2_beg[midpos].second;
			midpos++;
		}
		double med;
		if(w != .5 * totw)
			med = temp2_beg[midpos - 1].first;
		else
			med = .5 * (temp2_beg[midpos - 1].first + temp2_beg[midpos].first);

		double inv_yw_range = 1. / (yw_range_factor * med);
		for(int i = 0; i < npoints; i++)
		{
			w_beg[i] = y_weight_func((y_fit_beg[i] - y_beg[i]) * inv_yw_range) *
				weight_beg[i];
		}
	}
}

}

