// $Id: basefile_plugin.cc,v 1.5 2003/09/15 11:13:33 troein Exp $
//
// BioArray Software Environment (BASE) - homepage http://base.thep.lu.se/
// Copyright (C) 2002,2003 Lao Saal, Carl Troein, Johan Vallon-Christersson
//
// This file is part of BASE.
//
// BASE is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// BASE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
#ifdef HAVE_CONFIG_H
	#include <config.h>
#endif

#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <iostream>

#include <basefile_plugin.h>

namespace BASE
{

const std::string PluginBaseFileReader::_undefstr = "undefined";
const PluginBaseFileReader::Assay PluginBaseFileReader::_undefassay =
	{ 0, "undefined" };

const char *PluginBaseFileReader::_plerrtexts[] = 
{
	"Expected settings section", "Missing a setting in settings section",
	"Expected 'assays' section", "Malformatted 'assays' section",
	"No 'spots' section found in file", "Malformatted 'spots' section",
	"Improper 'spots' section for this plugin",
	"Version mismatch between plugin and data file",
	"Bug in plugin: bad sequence of function calls",
	"Bug in plugin: Bad argument in data retrival function"
};

PluginBaseFileReader::PluginBaseFileReader(std::istream &s):
	BaseFileReader(s)
{
}

PluginBaseFileReader::~PluginBaseFileReader()
{
}

const char *PluginBaseFileReader::errText(int e)
{
	if(e < MISSINGSETTINGS || e >= LASTERR)
		return BaseFileReader::errText(e);
	return _plerrtexts[e - MISSINGSETTINGS];
}


bool PluginBaseFileReader::registerSettings(
	const std::string &sectionname,
	const std::string &pluginversion,
	const strvec &s)
{
	_settings.clear();
	for(strvec::const_iterator i = s.begin(); i != s.end(); i++)
		_settings[*i] = "";
	_settingstype = sectionname;
	_pluginversion = pluginversion;
	return true;
}

const std::string &PluginBaseFileReader::getSetting(const std::string &name)
{
	return _settings[name];
}
int PluginBaseFileReader::getSettingInt(const std::string &name)
{
	return atoi(_settings[name].c_str());
}
double PluginBaseFileReader::getSettingDouble(const std::string &name)
{
	return atof(_settings[name].c_str());
}


bool PluginBaseFileReader::registerDataFormat(const strvec &columns,
	const strvec &fields, bool allIntensities, int minCh, int maxCh,
	format_t format)
{
	_reqcolumns = columns;
	_reqfields = fields;
	_reqallint = allIntensities;
	_reqserial = format == format_serial;
	_reqnormal = format == format_normal;
	_channelsmin = minCh;
	_channelsmax = maxCh;
	_datafmtregged = true;
	return true;
}


bool PluginBaseFileReader::readInfoSections()
{
	if(!isFirstSection())
	{
		_error = BADCALLSEQUENCE;
		_errdescr = "PluginBaseFileReader::readInfoSections() "
			"after reading was started";
		return false;
	}

	BaseFileSection sect;

	if(_settingstype != "")
	{
		if(!readSection(sect))
		{
			if(!_error)
				_error = MISSINGSETTINGS;
			return false;
		}
		if(sect.type != _settingstype)
		{
			_error = MISSINGSETTINGS;
			_errdescr = std::string("Missing section '") +
				_settingstype + "'";
			return false;
		}
		for(ssmap::iterator i = _settings.begin(); i != _settings.end(); i++)
		{
			BaseFileSection::header_t::iterator j =
				sect.headers.find((*i).first);
			if(j == sect.headers.end())
			{
				_error = BROKENSETTINGS;
				_errdescr = "Missing setting '" + (*i).first + "'";
				return false;
			}
			(*i).second = (*j).second;
		}
		_settings.swap(sect.headers);

		// Whine if the plugin definition is for the wrong version
		// of the plugin
		if(_pluginversion != "" &&
			getSetting("pluginVersion") != _pluginversion)
		{
			_error = BADVERSION;
			_errdescr = "Plugin executable is version " +
				_pluginversion + " but the input file is for "
				"version " + getSetting("pluginVersion");
			return false;
		}
	}

	if(!readSection(sect) || sect.type != "assays")
	{
		if(!_error)
			_error = MISSINGASSAYS;
		return false;
	}

	FieldList columns, annotationColumns;
	if(!sect.findFieldList("columns", columns) ||
		!sect.findFieldList("annotationColumns", annotationColumns))
	{
		_error = BROKENASSAYS;
		_errdescr = "Missing header 'columns' and/or 'annotationColumns'";
		return false;
	}
	int idCol, nameCol;
	if(!sect.findFieldPos("id", idCol, columns) ||
		!sect.findFieldPos("name", nameCol, columns))
	{
		_error = BROKENASSAYS;
		_errdescr = "Missing column 'id' and/or 'name'";
		return false;
	}

	for(FieldList::iterator i = annotationColumns.begin();
		i != annotationColumns.end(); i++)
	{
		_annottypes.push_back((*i).first);
		int acol;
		if(!sect.findFieldPos((*i).first.c_str(), acol, columns))
		{
			_error = BROKENASSAYS;
			_errdescr = "Missing an annotation column ('" +
				(*i).first + "') in 'columns'";
			return false;
		}
	}
	strvec cols(columns.size());
	while(readVec(cols))
	{
		Assay a;
		a.id = atoi(cols[idCol].c_str());
		a.name = cols[nameCol];
		for(FieldList::iterator i = annotationColumns.begin();
			i != annotationColumns.end(); i++)
		{
			a.annots.push_back(cols[columns[(*i).first]]);
		}
		_assaynums[a.id] = _assays.size();
		_assays.push_back(a);
	}
	if(getError())
		return false;

	return true;
}

bool PluginBaseFileReader::readSpotsSection()
{
	if(_assays.empty())
	{
		_error = BADCALLSEQUENCE;
		_errdescr = "PluginBaseFileReader::readSpotsSection() "
			"must be preceded by readInfoSections()";
		return false;
	}
	else if(!_datafmtregged)
	{
		_error = BADCALLSEQUENCE;
		_errdescr = "PluginBaseFileReader::readSpotsSection() "
			"must be preceded by registerDataFormat()";
		return false;
	}

	BaseFileSection sect;
	while(1)
	{
		if(!readSection(sect))
		{
			if(!_spotssection && _error == EOR)
				_error = MISSINGSPOTS;
			return false;
		}
		if(sect.type == "spots")
			break;
		std::cerr << "Warning: Unknown section of type '" << sect.type <<
			"' skipped in PluginBaseFileReader::readSpotsSection()\n";
	}
	_spotssection++;

	FieldList assays;
	if(!sect.findFieldList("columns", _datacolumns) ||
		!sect.findFieldList("assayFields", _datafields) ||
		!sect.findFieldList("assays", assays))
	{
		_error = BROKENSPOTS;
		_errdescr = "Missing header 'columns' or 'assayFields' or 'assays'";
		return false;
	}
	if(!BaseFileSection::findFieldPos("assayData", _assayDatacolumn,
		_datacolumns))
	{
		_error = BROKENSPOTS;
		_errdescr = "Missing column 'assayData'";
		return false;
	}

	if(_reqserial && assays.size() != 1)
	{
		_error = BADSPOTS;
		_errdescr = "Serial format (one assay per spots section) required";
		return false;
	}
	if(_reqnormal && assays.size() != _assays.size())
	{
		_error = BADSPOTS;
		_errdescr = "Normal format (all assays in one spots section) required";
		return false;
	}

	if(!sect.findIntOpt("channels", _channels))
	{
		_error = BROKENSPOTS;
		_errdescr = "Missing 'channels' header";
		return false;
	}

	if(_channelsmin > 0 && _channels < _channelsmin ||
		_channelsmax > 0 && _channels > _channelsmax)
	{
		_error = BADSPOTS;
		_errdescr = "The number of channels isn't in [" +
			toString(_channelsmin) + ", " +
			(_channelsmax > 0 ? toString(_channelsmax) : std::string("inf")) +
			"]";
		return false;
	}

	for(strvec::iterator i = _reqcolumns.begin(); i != _reqcolumns.end(); i++)
	{
		int col;
		if(!BaseFileSection::findFieldPos((*i).c_str(), col, _datacolumns))
		{
			_error = BADSPOTS;
			_errdescr = "Missing required column '" + (*i) + "'";
			return false;
		}
	}
	for(strvec::iterator i = _reqfields.begin(); i != _reqfields.end(); i++)
	{
		int field;
		if(!BaseFileSection::findFieldPos((*i).c_str(), field, _datafields))
		{
			_error = BADSPOTS;
			_errdescr = "Missing required field '" + (*i) + "'";
			return false;
		}
	}
	if(_reqallint)
	{
		for(int ch = 1; ch <= _channels; ch++)
		{
			int field;
			std::string fname = std::string("intensity") + toString(ch);
			if(!BaseFileSection::findFieldPos(fname.c_str(), field, _datafields))
			{
				_error = BADSPOTS;
				_errdescr = "Missing required intensity field '" + fname + "'";
				return false;
			}
		}
	}

	_dataassays.clear();
	for(FieldList::iterator i = assays.begin(); i != assays.end(); i++)
	{
		int aid = atoi((*i).first.c_str());
		std::map<int, int>::iterator j = _assaynums.find(aid);
		if(j == _assaynums.end())
		{
			_error = BADSPOTS;
			_errdescr = "Reference to unknown assay '" + (*i).first + "'";
			return false;
		}
		_dataassays.push_back((*j).second);
	}

	_intensityfields.resize(_channels);
	for(int ch = 1; ch <= _channels; ch++)
	{
		simap::iterator i = _datafields.find(
			std::string("intensity") + toString(ch));
		if(i == _datafields.end())
			_intensityfields[ch - 1] = -1;
		else
			_intensityfields[ch - 1] = (*i).second;
	}

	_databuf.resize(_datacolumns.size() - 1 +
		_dataassays.size() * _datafields.size());

	return true;
}

const PluginBaseFileReader::Assay &PluginBaseFileReader::getAssay(int assay)
{
	if(assay < 0 || assay >= (int)_dataassays.size())
	{
		_error = BADINDEXARGUMENT;
		_errdescr = "Arg to PluginBaseFileReader::getAssay() outside range: " +
			toString(assay) + " not in [0, " + toString(_dataassays.size()) +
			")";
		return _undefassay;
	}
	return _assays[_dataassays[assay]];
}

const std::string &PluginBaseFileReader::getAnnotationName(int annot)
{
	if(annot < 0 || annot >= (int)_annottypes.size())
	{
		_error = BADINDEXARGUMENT;
		_errdescr = "Arg to PluginBaseFileReader::getAnnotationName() "
			"outside range: " + toString(annot) + " not in [0, " +
			toString(_annottypes.size())  + ")";
		return _undefstr;
	}
	return _annottypes[annot];
}


bool PluginBaseFileReader::readDataLine()
{
	return readVec(_databuf);
}

const char *PluginBaseFileReader::getColumn(const std::string &name)
{
	simap::iterator i = _datacolumns.find(name);
	if(i == _datacolumns.end())
	{
		_error = BADINDEXARGUMENT;
		_errdescr = "Bad arg to PluginBaseFileReader::getColumn(): '" +
			name + "' is not a column in this file";
		return _undefstr.c_str();
	}
	return _databuf[(*i).second];
}
std::string PluginBaseFileReader::getColumnString(const std::string &name)
{
	return std::string(getColumn(name));
}
int PluginBaseFileReader::getColumnInt(const std::string &name)
{
	return atoi(getColumn(name));
}
double PluginBaseFileReader::getColumnDouble(const std::string &name)
{
	const char *d = getColumn(name);
	if(!*d)
		return dnan;
	return atof(d);
}

const char *PluginBaseFileReader::getField(const std::string &name,
	int assay)
{
	if(assay < 0 || assay >= (int)_dataassays.size())
	{
		_error = BADINDEXARGUMENT;
		_errdescr = "Bad assay number (" + toString(assay) +
			") in PluginBaseFileReader::getField()";
		return _undefstr.c_str();
	}
	simap::iterator i = _datafields.find(name);
	if(i == _datafields.end())
	{
		_error = BADINDEXARGUMENT;
		_errdescr = "Undefined field '" + name +
			"' in PluginBaseFileReader::getField()";
		return _undefstr.c_str();
	}
	return _databuf[(*i).second + _assayDatacolumn +
		assay * _datafields.size()];
}
std::string PluginBaseFileReader::getFieldString(const std::string &name,
	int assay)
{
	return std::string(getField(name, assay));
}
int PluginBaseFileReader::getFieldInt(const std::string &name,
	int assay)
{
	return atoi(getField(name, assay));
}
double PluginBaseFileReader::getFieldDouble(const std::string &name,
	int assay)
{
	const char *d = getField(name, assay);
	if(!*d)
		return dnan;
	return atof(d);
}

double PluginBaseFileReader::getIntensity(int assay, int channel)
{
	if(assay < 0 || assay >= (int)_dataassays.size())
	{
		_error = BADINDEXARGUMENT;
		_errdescr = "Bad assay number (" + toString(assay) +
			") in PluginBaseFileReader::getIntensity()";
		return dnan;
	}
	if(channel < 1 || channel > _channels)
	{
		_error = BADINDEXARGUMENT;
		_errdescr = "Bad channel number (" + toString(channel) +
			") in PluginBaseFileReader::getIntensity()";
		return dnan;
	}
	if(_intensityfields[channel - 1] < 0)
	{
		_error = BADINDEXARGUMENT;
		_errdescr = "'intensity" + toString(channel) +
			"' not in field list in PluginBaseFileReader::getIntensity()";
		return dnan;
	}
	const char *d = _databuf[_intensityfields[channel - 1] +
		_assayDatacolumn + assay * _datafields.size()];
	if(!*d)
		return dnan;
	return atof(d);
}


std::string PluginBaseFileReader::toString(int i)
{
	static char buf[32];
	std::sprintf(buf, "%d", i);
	return std::string(buf);
}


};
